import typing

import hashlib
import x25519

from nacl.signing import SigningKey as ed25519Private, VerifyKey as ed25519Public
from nacl.public import PublicKey as x25519Public, PrivateKey as x25519Private

from Cryptodome.Random import get_random_bytes
from Cryptodome.Cipher import AES


class Crypto:
    magic = b'\xc6\xb4\x13\x48'
    magic_aes = b'\xd4\xad\xbc-'

    def __init__(self):
        raise NotImplementedError

    def get_key_id(self):
        return hashlib.sha256(self.magic + self.ed25519_public.encode()).digest()

    def get_aes_key_id(self):
        return hashlib.sha256(self.magic_aes + self.ed25519_private.encode()).digest()


class Server(Crypto):

    def __init__(self, host: str, port: int, pub_key: bytes):
        self.host = host
        self.port = port
        self.ed25519_public = ed25519Public(pub_key)
        self.x25519_public = self.ed25519_public.to_curve25519_public_key()


class Client(Crypto):
    def __init__(self, ed25519_private_key: bytes) -> None:
        self.ed25519_private = ed25519Private(ed25519_private_key)
        self.ed25519_public = self.ed25519_private.verify_key
        self.x25519_private = self.ed25519_private.to_curve25519_private_key()
        self.x25519_public = self.x25519_private.public_key

    @staticmethod
    def generate_ed25519_private_key() -> bytes:
        return bytes(ed25519Private.generate())

    def sign(self, message: bytes) -> bytes:
        return get_signature(self.ed25519_private, message)


class AdnlChannel:
    """
    A class that represents adnl channel with node
    """

    def __init__(self, client: Client, server: Server, local_id: bytes, peer_id: bytes):
        self.client_channel = client
        self.server_channel = server
        self.channel_shared = get_shared_key(self.client_channel.x25519_private.encode(), self.server_channel.x25519_public.encode())
        if local_id > peer_id:
            self.enc_key = self.channel_shared
            self.dec_key = self.channel_shared[::-1]
        elif local_id < peer_id:
            self.enc_key = self.channel_shared[::-1]
            self.dec_key = self.channel_shared
        else:
            self.enc_key = self.channel_shared
            self.dec_key = self.channel_shared
        self.client_aes_key_id = get_key_aes_id(self.enc_key)
        self.server_aes_key_id = get_key_aes_id(self.dec_key)

    def encrypt(self, data: bytes) -> bytes:
        checksum = hashlib.sha256(data).digest()
        enc_cipher = create_aes_ctr_sipher_from_key_n_data(self.enc_key, checksum)
        data = aes_ctr_encrypt(enc_cipher, data)
        return self.client_aes_key_id + checksum + data

    def decrypt(self, encrypted_data: bytes, checksum: bytes) -> bytes:
        dec_cipher = create_aes_ctr_sipher_from_key_n_data(self.dec_key, checksum)
        return aes_ctr_decrypt(dec_cipher, encrypted_data)


def get_random(bytes_size: int) -> bytes:
    return get_random_bytes(bytes_size)


def create_aes_ctr_sipher_from_key_n_data(key: bytes, data: bytes):
    return create_aes_ctr_cipher(key[0:16] + data[16:32], data[0:4] + key[20:32])


def create_aes_ctr_cipher(key: bytes, iv: bytes):
    if len(key) != 32:
        raise Exception('key should be 32 bytes exactly!')

    cipher = AES.new(key, AES.MODE_CTR, initial_value=iv, nonce=b'')
    # cipher = Cipher(algorithms.AES(key), modes.CTR(iv)).encryptor()

    return cipher


def get_key_aes_id(key: bytes):
    return hashlib.sha256(b'\xd4\xad\xbc-' + key).digest()


def aes_ctr_encrypt(cipher, data: bytes) -> bytes:
    # return cipher.encryptor().update(data)
    return cipher.encrypt(data)


def aes_ctr_decrypt(cipher, data: bytes) -> bytes:
    # return cipher.decryptor().update(data)
    return cipher.decrypt(data)


def get_shared_key(private_key: bytes, public_key: bytes) -> bytes:
    """
    :param public_key: peer public x25519 key
    :param private_key: client private x25519 key
    :return: ECDH x25519 shared key
    """
    return x25519.scalar_mult(private_key, public_key)


def get_signature(private_key: ed25519Private, message: bytes) -> bytes:
    return ed25519Private(seed=private_key.encode()).sign(message)[:64]



import typing

from ..tlb import TlbScheme
from ...boc import Cell, Builder, Slice, HashMap, Address


class NftItemData(TlbScheme):
    """
    nft_item_data#_ index:uint64 collection_address:Address owner_address:Address content:^Cell = NftItemData;
    """
    def __init__(self,
                 index: typing.Optional[int] = 0,
                 collection_address: typing.Optional[Address] = None,
                 owner_address: typing.Optional[Address] = None,
                 content: typing.Optional[Cell] = None
                 ):
        self.index = index
        if isinstance(collection_address, str):
            collection_address = Address(collection_address)
        if isinstance(owner_address, str):
            collection_address = Address(collection_address)
        self.collection_address = collection_address
        self.owner_address = owner_address
        self.content = content

    def serialize(self) -> Cell:
        builder = Builder()
        builder\
            .store_uint(self.index, 64)\
            .store_address(self.collection_address)\
            .store_address(self.owner_address)\
            .store_ref(self.content)
        return builder.end_cell()

    @classmethod
    def deserialize(cls, cell_slice: Slice):
        return cls(index=cell_slice.load_uint(64), collection_address=cell_slice.load_address(), owner_address=cell_slice.load_address(), content=cell_slice.load_ref())


class NftItemSaleFees(TlbScheme):
    """
    nft_item_sale_fees#_ marketplace_fee_address:Address marketplace_fee:Grams royalty_address:Address royalty_amount:Grams = NftItemSaleFees;
    """
    def __init__(self,
                 marketplace_fee_address: typing.Optional[Address] = None,
                 marketplace_fee: typing.Optional[int] = None,
                 royalty_address: typing.Optional[Address] = None,
                 royalty_amount: typing.Optional[int] = None
                 ):
        self.marketplace_fee_address = marketplace_fee_address
        self.marketplace_fee = marketplace_fee
        self.royalty_address = royalty_address
        self.royalty_amount = royalty_amount

    def serialize(self):
        return Builder()\
            .store_address(self.marketplace_fee_address)\
            .store_coins(self.marketplace_fee)\
            .store_address(self.royalty_address)\
            .store_coins(self.royalty_amount)\
            .end_cell()

    @classmethod
    def deserialize(cls, cell_slice: Slice):
        return cls(
            marketplace_fee_address=cell_slice.load_address(),
            marketplace_fee=cell_slice.load_coins(),
            royalty_address=cell_slice.load_address(),
            royalty_amount=cell_slice.load_coins()
        )


class NftItemSaleData(TlbScheme):
    """
    nft_item_sale_data#_ is_complete:bool created_at:uint32 marketplace_address:Address nft_address:Address nft_owner_address:Address full_price:grams fees_cell:^NftItemSaleFees can_deploy_by_external:bool = NftItemSaleData;
    """
    def __init__(self,
                 is_complete: typing.Optional[bool] = None,
                 created_at: typing.Optional[int] = None,
                 marketplace_address: typing.Optional[Address] = None,
                 nft_address: typing.Optional[Address] = None,
                 nft_owner_address: typing.Optional[Address] = None,
                 full_price: typing.Optional[int] = None,
                 fees_cell: typing.Optional[NftItemSaleFees] = None,
                 can_deploy_by_external: typing.Optional[bool] = None,
                 ):
        self.is_complete = is_complete
        self.created_at = created_at
        if isinstance(marketplace_address, str):
            marketplace_address = Address(marketplace_address)
        if isinstance(nft_address, str):
            nft_address = Address(nft_address)
        if isinstance(nft_owner_address, str):
            nft_owner_address = Address(nft_owner_address)
        self.marketplace_address = marketplace_address
        self.nft_address = nft_address
        self.nft_owner_address = nft_owner_address
        self.full_price = full_price
        self.fees_cell = fees_cell
        self.can_deploy_by_external = can_deploy_by_external

    def serialize(self) -> Cell:
        builder = Builder()
        builder\
            .store_bool(self.is_complete)\
            .store_uint(self.created_at, 32)\
            .store_address(self.marketplace_address)\
            .store_address(self.nft_address)\
            .store_address(self.nft_owner_address)\
            .store_coins(self.full_price)\
            .store_ref(self.fees_cell.serialize())\
            .store_bool(self.can_deploy_by_external)
        return builder.end_cell()

    @classmethod
    def deserialize(cls, cell_slice: Slice) -> "NftItemSaleData":
        return cls(
            is_complete=cell_slice.load_bool(),
            created_at=cell_slice.load_uint(32),
            marketplace_address=cell_slice.load_address(),
            nft_address=cell_slice.load_address(),
            nft_owner_address=cell_slice.load_address(),
            full_price=cell_slice.load_coins(),
            fees_cell=NftItemSaleFees.deserialize(cell_slice.load_ref().begin_parse()),
            can_deploy_by_external=cell_slice.load_bool()
        )





import typing


def make_url_safe(str: str) -> str:
    return str.replace("+", '-').replace("/", '_')


def bytes_to_uint(data: bytes):
    return int.from_bytes(data, 'big', signed=False)


def bin_to_signed_int(ba: str) -> int:
    from bitstring import BitArray
    return BitArray(bin=ba).int



import typing


class BlockId:
    """
    tonNode.blockId workchain:int shard:long seqno:int = tonNode.BlockId;
    """
    def __init__(self, workchain: int, shard: typing.Optional[int], seqno: int):
        if shard is None:
            shard = -9223372036854775808
        self.workchain = workchain
        self.shard = shard
        self.seqno = seqno

    def to_dict(self):
        return {'workchain': self.workchain, 'shard': self.shard, 'seqno': self.seqno}

    @classmethod
    def from_dict(cls, block_id: dict):
        return cls(workchain=block_id.get('workchain'), shard=block_id.get('shard'), seqno=block_id.get('seqno'))


class BlockIdExt:
    """
    tonNode.blockIdExt workchain:int shard:long seqno:int root_hash:int256 file_hash:int256 = tonNode.BlockIdExt;
    """

    def __init__(self, workchain: int, shard: typing.Optional[int], seqno: int,
                 root_hash: typing.Union[str, bytes], file_hash: typing.Union[str, bytes]):
        if shard is None:
            shard = -9223372036854775808
        if isinstance(root_hash, str):
            root_hash = bytes.fromhex(root_hash)
        if isinstance(file_hash, str):
            file_hash = bytes.fromhex(file_hash)
        self.workchain = workchain
        self.shard = shard
        self.seqno = seqno
        self.root_hash = root_hash
        self.file_hash = file_hash

    def to_dict(self):
        return {'workchain': self.workchain, 'shard': self.shard, 'seqno': self.seqno, 'root_hash': self.root_hash.hex(), 'file_hash': self.file_hash.hex()}

    @classmethod
    def from_dict(cls, block_id_ext: dict):
        return cls(workchain=block_id_ext.get('workchain'), shard=block_id_ext.get('shard'), seqno=block_id_ext.get('seqno'),
                   root_hash=block_id_ext.get('root_hash'), file_hash=block_id_ext.get('file_hash'))

    def to_bytes(self) -> bytes:
        return self.workchain.to_bytes(4, 'big', signed=True) + self.shard.to_bytes(8, 'big', signed=True) + self.seqno.to_bytes(4, 'big', signed=True) + self.root_hash + self.file_hash

    @classmethod
    def from_bytes(cls, data: bytes) -> "BlockIdExt":
        wc = int.from_bytes(data[:4], 'big', signed=True)
        shard = int.from_bytes(data[4:12], 'big', signed=True)
        seqno = int.from_bytes(data[12:16], 'big', signed=True)
        root_hash = data[16:48]
        file_hash = data[48:80]
        return cls(workchain=wc, shard=shard, seqno=seqno, root_hash=root_hash, file_hash=file_hash)

    def __repr__(self):
        return f'<TL BlockIdExt [wc={self.workchain}, shard={self.shard}, seqno={self.seqno}, root_hash={self.root_hash.hex()}, file_hash={self.file_hash.hex()}] >'

    def __eq__(self, other: "BlockIdExt"):
        if self.seqno != other.seqno or self.workchain != other.workchain or self.shard != other.shard or self.root_hash != other.root_hash or self.file_hash != other.file_hash:
            return False
        return True

    def __hash__(self):
        return self.root_hash



from typing import Union, Iterable, overload

from bitarray import bitarray
from bitarray.util import int2ba


BytesLike = Union[bytes, Iterable[int]]


class TvmBitarrayException(Exception):
    pass


class TvmBitarrayOverflowException(TvmBitarrayException):
    pass


class TvmBitarrayUnderflowException(TvmBitarrayException):
    pass


class TvmBitarray(bitarray):

    def __new__(cls, size: int = 1023, *args, **kwargs):
        return super().__new__(cls, *args, **kwargs)

    def __init__(self, size: int = 1023, *args, **kwargs):
        if size > 1023:
            raise TvmBitarrayException('bitarray size must be <= 1023')
        self._size = size
        super().__init__()

    def check_overflow(self, length: int) -> None:
        if len(self) + length > 1023:
            raise TvmBitarrayOverflowException('bitstring overflow')

    def check_underflow(self, length: int) -> None:
        if len(self) < length:
            raise TvmBitarrayUnderflowException('bitstring underflow')

    def extend(self, x: Union[str, Iterable[int]]) -> None:
        self.check_overflow(len(x))
        super().extend(x)

    def append(self, value: int) -> None:
        self.check_overflow(1)
        super().append(value)

    def frombytes(self, a: BytesLike) -> None:
        self.check_overflow(len(a) * 8)
        super().frombytes(a)

    def copy(self) -> "TvmBitarray":
        res = self.__new__(TvmBitarray)
        res.extend(self)
        return res

    def __delitem__(self, item: Union[int, slice]):
        if isinstance(item, slice):
            start = item.start if item.start else 0
            stop = item.stop if item.stop else len(self)
            self.check_underflow(stop - start)
        elif isinstance(item, int):
            self.check_underflow(1)
        return super().__delitem__(item)

    def to_bitarray(self):
        return bitarray(self)


BitarrayLike = Union[TvmBitarray, bitarray]



import typing

from .tlb import TlbScheme, TlbError
from .. import Builder
from ..boc import Slice, Cell, CellTypes


class MerkleUpdate(TlbScheme):
    """
    !merkle_update#02 {X:Type} old_hash:bits256 new_hash:bits256 old:^X new:^X = MERKLE_UPDATE X;
    """

    def __init__(self, cell: Cell, old_hash: bytes, new_hash: bytes, old, new):
        self.cell = cell
        self.old_hash = old_hash
        self.new_hash = new_hash
        self.old = old
        self.new = new

    @classmethod
    def serialize(cls, *args): ...

    @classmethod
    def deserialize(cls, cell: Cell, deserializer: typing.Callable) -> typing.Optional["MerkleUpdate"]:
        if cell.type_ != CellTypes.merkle_update:
            return None

        cell_slice = cell.begin_parse()
        tag = cell_slice.load_bytes(1)[:1]
        # if tag != b'\x02':
        #     raise TlbError(f'MerkleUpdate deserialization error: unexpected tag {tag}')
        old_hash = cell_slice.load_bytes(32)
        new_hash = cell_slice.load_bytes(32)
        old = deserializer(cell_slice.load_ref().begin_parse())
        new = deserializer(cell_slice.load_ref().begin_parse())
        return cls(cell, old_hash, new_hash, old, new)


class HashUpdate(TlbScheme):
    """
    update_hashes#72 {X:Type} old_hash:bits256 new_hash:bits256 = HASH_UPDATE X;
    """

    def __init__(self, old_hash: bytes, new_hash: bytes):
        self.old_hash = old_hash
        self.new_hash = new_hash

    def serialize(self):
        return Builder()\
                .store_bytes(b'\x72')\
                .store_bytes(self.old_hash)\
                .store_bytes(self.new_hash)\
                .end_cell()

    @classmethod
    def deserialize(cls, cell_slice: Slice):
        tag = cell_slice.load_bytes(1)[:1]
        if tag != b'r':
            raise TlbError(f'HashUpdate deserialization error: unexpected tag {tag}')
        old_hash = cell_slice.load_bytes(32)
        new_hash = cell_slice.load_bytes(32)
        return cls(old_hash, new_hash)


def deserialize_shard_hashes(cell_slice: Slice):
    from .block import BinTree, ShardDescr
    shard_hashes = cell_slice.load_dict(32, value_deserializer=lambda src: BinTree.deserialize(
        src.load_ref().begin_parse()))
    if shard_hashes:
        for k in shard_hashes:
            for i in range(len(shard_hashes[k].list)):
                shard_hashes[k].list[i] = ShardDescr.deserialize(shard_hashes[k].list[i])
    return shard_hashes


def uint64_to_int64(num: int):
    return (num & ((1 << 63) - 1)) - (num & (1 << 63))



import hashlib
import typing


from ..tlb.block import Block, ShardStateUnsplit
from ..tlb.config import ValidatorDescr, CatchainConfig, ValidatorSet
from ..tl.block import BlockId, BlockIdExt
from ..crypto.signature import verify_sign
from ..boc.tvm_bitarray import TvmBitarray
from ..boc.exotic import CellTypes
from ..boc.cell import Cell
from ..boc.address import Address


class ProofError(Exception):
    pass


def check_proof(cell: "Cell", hash_: bytes) -> None:
    if cell.type_ != CellTypes.merkle_proof:
        raise ProofError(f'Expected Merkle proof Cell, got {cell.type_} Cell type')

    if cell.data[1:33] != hash_:
        raise ProofError('Provided invalid hash')

    if cell[0].get_hash(0) != hash_:  # TODO Level mask
        raise ProofError('Merkle proof is invalid')

    return


# https://github.com/ton-blockchain/ton/blob/master/crypto/block/check-proof.cpp
def check_block_header_proof(root_cell: "Cell", block_hash: bytes, store_state_hash=False):
    root_hash = root_cell.get_hash(0)
    if root_hash != block_hash:
        raise ProofError('Block header proof error: hashes unmatch')
    if store_state_hash:
        state_update = root_cell[2][1]
        return state_update.get_hash(0)
    return


def check_shard_proof(shard_proof: bytes, blk: BlockIdExt, shrd_blk: BlockIdExt):

    if blk == shrd_blk:
        return
    if blk.workchain != -1:
        raise ProofError('expected masterchain block')
    shard_proof_cells = Cell.from_boc(shard_proof)

    if len(shard_proof_cells) != 2:
        raise ProofError('expected 2 root cells in shard proof')

    mc_block_cell = shard_proof_cells[0]
    mc_state_root = shard_proof_cells[1]

    block_info = Block.deserialize(mc_block_cell[0].begin_parse()).info

    if not (block_info.seqno == blk.seqno and block_info.shard.workchain_id == blk.workchain):
        raise ProofError('block info mismatch')

    mc_state_hash = mc_state_root[0].get_hash(0)
    state_hash = check_block_header_proof(mc_block_cell[0], blk.root_hash, True)

    if mc_state_hash != state_hash:
        raise ProofError('mc state hashes mismatch')

    shard = ShardStateUnsplit.deserialize(mc_state_root[0].begin_parse())

    shard_dict = shard.custom.shard_hashes

    shard_descr = shard_dict.get(shrd_blk.workchain)
    if shard_descr is None:
        raise ProofError('cannot find shard block in ShardHashes')

    # shard_dict looks like this:
    # {0: {'list': [{'seq_no': 36667578, 'reg_mc_seqno': 30985767, 'start_lt': 39143950000000, 'end_lt': 39143950000004, 'root_hash': b'\x83f\xe4P\x05\xf3\xbe\x83\xaf\x00\xf9\xbf\xde\x01\x9e=fF%\xd4\xed&\xb9\x0b\xd4\n\xb1]K\xedd9', 'file_hash': b"n\xdd\xdb\xc8\xb01\xbe\xcf\xf7\xed\xb0'3\x9b\t\xe7\xd7J\xac/\x14W!y\x82o\xf6\x19\xeb\xeaj\xae", 'before_split': False, 'before_merge': False, 'want_split': False, 'want_merge': True, 'nx_cc_updated': False, 'flags': 0, 'next_catchain_seqno': 456373, 'next_validator_shard': 9223372036854775808, 'min_ref_mc_seqno': 30985764, 'gen_utime': 1688896742, 'split_merge_at': None, 'fees_collected': {'grams': 1006293421, 'other': {'dict': None}}, 'funds_created': {'grams': 1000000000, 'other': {'dict': None}}}]}}

    shard_descr = shard_descr.list[0]  # see BinTree TLB schema

    if shard_descr.root_hash != shrd_blk.root_hash:
        raise ProofError('shard block actual and expected hashes mismatch')

    return shard_descr


def check_account_proof(proof: bytes, shrd_blk: BlockIdExt, address: "Address", account_state_root: "Cell", return_account_descr: bool = False):

    proof_cells = Cell.from_boc(proof)
    if len(proof_cells) != 2:
        raise ProofError('expected 2 root cells in account state proof')

    state_cell = proof_cells[1]

    state_hash = check_block_header_proof(proof_cells[0][0], shrd_blk.root_hash, True)

    if state_cell[0].get_hash(0) != state_hash:
        raise ProofError('state hashes mismatch')

    shard = ShardStateUnsplit.deserialize(state_cell[0].begin_parse())

    shard_account = shard.accounts[0][int.from_bytes(address.hash_part, 'big')]

    account_state_root_proved = shard_account.cell

    if account_state_root_proved[0].get_hash(0) != account_state_root.get_hash(0):
        raise ProofError('account state proof invalid')

    if return_account_descr:
        return shard_account

    return


def calculate_node_id_short(pub_key: bytes):
    return hashlib.sha256(b'\xc6\xb4\x13H' + pub_key).digest()


def check_block_signatures(nodes: typing.List[ValidatorDescr], signatures: typing.List[dict], blk: BlockIdExt):
    from ..tlb.config import ValidatorDescr

    node_map = {}

    total_weight = 0
    signed_weight = 0
    i = 0
    for node in nodes:
        total_weight += node.weight
        i += 1

        node_map[calculate_node_id_short(node.public_key.pubkey)] = node

    to_sign = b'pn\x0b\xc5' + blk.root_hash + blk.file_hash  # bytes.fromhex('c50b6e70')[::-1] - magic
    i = 0
    for sig in signatures:
        node = node_map.get(bytes.fromhex(sig['node_id_short']))
        node: ValidatorDescr
        i += 1

        if node is None:
            raise ProofError('cannot find node_id_short in validator list')

        result = verify_sign(public_key=node.public_key.pubkey, signed_message=to_sign, signature=sig['signature'])

        if not result:
            raise ProofError('invalid signature!')

        signed_weight += node.weight

    if signed_weight * 3 >= total_weight * 2:  # >= 2/3
        return

    raise ProofError(f'Block {blk} has not been signed by 2/3 of validators')


def compute_validator_set(ccv_conf: CatchainConfig, blk: BlockIdExt, vset: ValidatorSet, cc_seqno: int = None):
    is_mc = blk.workchain == -1

    if is_mc:
        count = vset.main
    else:
        count = ccv_conf.shard_validators_num
    count = min(count, vset.total)

    if is_mc:
        if ccv_conf.shuffle_mc_validators:
            # TODO
            pass
        return list(vset.list.values())[:count]
    # TODO true algorithm
    nodes = []
    for i in range(count):
        node = vset.list[i + vset.total]
        node.weight = 1  # shardchain validator lists have all weights = 1 ?
        nodes.append(node)
    return nodes



import pytest

from pytoniq_core.boc import Builder, begin_cell, Slice, Address


def test_boc():

    empty_cs = Slice.one_from_boc('b5ee9c72010101010002000000')

    assert empty_cs.remaining_bits == 0
    assert empty_cs.remaining_refs == 0


def test_bits():

    for btcnt in range(0, 1023):
        cs = begin_cell().store_bits('1' * btcnt).end_cell().begin_parse()
        assert cs.remaining_bits == btcnt
        assert cs.load_bits(btcnt).to01() == '1' * btcnt


def test_preload():

    cs = begin_cell().store_int(-100, 100).end_cell().begin_parse()

    cs.skip_bits(15)

    assert cs.preload_int(85) == -100
    assert cs.remaining_bits == 85


@pytest.mark.parametrize("addr", [
    'EQDtFpEwcFAEcRe5mLVh2N6C0x-_hJEM7W61_JLnSF74p4q2',
    None,
    '0:ed1691307050047117b998b561d8de82d31fbf84910ced6eb5fc92e7485ef8a7',
    '-1:ed1691307050047117b998b561d8de82d31fbf84910ced6eb5fc92e7485ef8a7',
    (0, b'\xed\x16\x910pP\x04q\x17\xb9\x98\xb5a\xd8\xde\x82\xd3\x1f\xbf\x84\x91\x0c\xedn\xb5\xfc\x92\xe7H^\xf8\xa7'),
    'Ef_tFpEwcFAEcRe5mLVh2N6C0x-_hJEM7W61_JLnSF74p3X-',
    'EQHtFpEwcFAEcRe5mLVh2N6C0x-_hJEM7W61_JLnSF74pwdq'
])
def test_address(addr):
    if isinstance(addr, tuple):
        cs = begin_cell().store_address(Address(addr)).end_cell().begin_parse()
    else:
        cs = begin_cell().store_address(addr).end_cell().begin_parse()

    if addr is None:
        assert cs.preload_address() is None
        assert cs.load_address() is None
        return

    assert cs.preload_address() == Address(addr)
    assert cs.load_address() == Address(addr)


def test_var_ints():

    assert begin_cell().store_var_int(-10, 10).to_slice().load_var_int(10) == -10
    assert begin_cell().store_var_uint(10, 10).to_slice().load_var_uint(10) == 10




import base64
import binascii

from ..crypto.crc import crc16
from .cell import Cell


class AddressError(Exception):
    pass


class Address:

    def __init__(self, address):
        """
        Note that initializing from hex form is 10 times faster than user-friendly bounceable form
        """
        self.wc: int = None
        self.hash_part: bytes = None
        self.is_bounceable = False
        self.is_test_only = False

        if isinstance(address, tuple):
            # Address((-1, b'\x11\x01\xff...'))
            self.wc = address[0]
            assert isinstance(address[1], bytes), 'expected bytes address hash part'
            self.hash_part = address[1]
            return
        if isinstance(address, self.__class__):
            self.wc = address.wc
            self.hash_part = address.hash_part
            return
        if self.is_hex(address):
            return
        if self.is_b64(address):
            return

        raise AddressError('unknown address type provided')

    def is_hex(self, addr: str) -> bool:
        try:
            wc, hash_part = addr.split(':')
            int(hash_part, 16)
            self.wc = int(wc)
            self.hash_part = bytes.fromhex(hash_part)
            return True
        except ValueError:
            return False

    def is_b64(self, addr: str) -> bool:
        try:
            decoded = base64.urlsafe_b64decode(addr)
            tag = decoded[0]
            if tag & 0x80:  # test flag
                self.is_test_only = True
                tag ^= 0x80
            if tag == 0x11:  # bounceable
                self.is_bounceable = True
            self.wc = int.from_bytes(decoded[1:2], 'big', signed=True)
            self.hash_part = decoded[2:34]
            if decoded[34:] != crc16(decoded[:34]):
                raise AddressError('the address is invalid')
            return True
        except binascii.Error:
            return False

    def to_str(self, is_user_friendly=True, is_url_safe=True, is_bounceable=True, is_test_only=False):
        """
        Note that to_str(is_user_friendly=False) is 20 times faster than to_str(is_user_friendly=True)
        """
        # interface reference https://github.com/tonfactory/tonsdk/blob/master/tonsdk/utils/_address.py#L108

        if not is_user_friendly:
            return f'{self.wc}:{self.hash_part.hex()}'

        tag = 0x11  # bounceable tag

        if not is_bounceable:
            tag = 0x51
        if is_test_only:
            tag |= 0x80

        result = tag.to_bytes(1, 'big') + self.wc.to_bytes(1, 'big', signed=True) + self.hash_part

        result += crc16(result)

        if is_url_safe:
            result = base64.urlsafe_b64encode(result).decode()
        else:
            result = base64.b64encode(result).decode()

        return result

    def to_tl_account_id(self) -> dict:
        return {'workchain': self.wc, 'id': self.hash_part.hex()}

    def to_cell(self) -> Cell:
        from .builder import Builder
        return Builder()\
            .store_bits('100')\
            .store_int(self.wc, 8)\
            .store_bytes(self.hash_part)\
            .end_cell()

    @classmethod
    def from_tonsdk_address(cls, address):
        """
        Usage: Address.from_tonsdk_address(tonsdk.utils.Address('address'))
        """
        return cls(address.to_string(False))

    def to_tonsdk_address(self, cls):
        """
        Usage: Address('0:33333...').to_tonsdk_address(tonsdk.utils.Address)
        """
        return cls(self.to_str(False))

    # def __str__(self):
    #     return self.to_str()

    def __eq__(self, other: "Address"):
        return self.wc == other.wc and self.hash_part == other.hash_part

    def __repr__(self):
        return f'Address<{self.to_str()}>'
    
    def __hash__(self):
        return int.from_bytes(self.hash_part, "big") + self.wc



import logging
import re
import zlib
import typing
import os


logger = logging.getLogger(name='TL')


class TlError(Exception):
    pass


class TlSchema:

    def __init__(self, id: typing.Optional[bytes], name: typing.Optional[str], class_name: typing.Optional[str], args: dict) -> None:
        self._id: bytes = id
        self._name: str = name
        self._class_name: str = class_name
        self._args: typing.Dict[str, str] = args  # {'param_name': 'param_type'}

    @property
    def id(self) -> bytes:
        return self._id

    def little_id(self) -> bytes:
        return self._id[::-1]

    @property
    def name(self) -> str:
        return self._name

    @property
    def class_name(self) -> str:
        return self._class_name

    @property
    def boxed(self) -> bool:
        return self._boxed

    @boxed.setter
    def boxed(self, is_boxed: bool):
        self._boxed = is_boxed

    @property
    def args(self) -> typing.Dict[str, str]:
        return self._args

    def is_empty(self) -> bool:
        return True if not self.id or not self.name else False

    @classmethod
    def empty(cls) -> "TlSchema":
        return cls(None, None, None, {})

    def __repr__(self) -> str:
        return f'TL Schema {self._name} №{self._id.hex()} with args {self._args}'
        # return f'TL Schema {self._name} №{self._id.hex()}'


class TlSchemas:

    base_types = {  # bytes len
        'Bool': 4,
        '#': 4,
        'int': 4,
        'long': 8,
        'int128': 16,
        'int256': 32,
        'string': None,
        'bytes': None,
        'vector': None,
    }

    def __init__(self, schemas: typing.List[TlSchema], auto_deserialize: bool = True):
        """
        :param schemas: TL Schemes
        :param auto_deserialize: if true will try to deserialize bytes fields with known TL schemes but skip it for schemes that are "untouchables"
        """
        self.list: list = schemas

        self.id_map: typing.Dict[bytes, TlSchema] = {}
        self.name_map: typing.Dict[str, TlSchema] = {}
        self.class_name_map: typing.Dict[str, typing.List[TlSchema]] = {}
        self.generate_map()

        self._auto_deserialize = auto_deserialize

        self.untouchables: typing.Dict[str, set] = {}  # {schema_name: {field1, field2}}
        self.set_default_untouchables()
        #  schemas and their fields not to auto deserialize. In ADNL some schemes (like adnl.message.part or rldp ones)
        #  are not supposed to be auto deserialized, because it may cause some errors.

    def get_by_id(self, tl_id: typing.Union[bytes, int], byteorder: typing.Literal['little', 'big'] = 'big') -> TlSchema:
        """
        :param tl_id: id of TL schema
        :param byteorder: provided bytes or int order. if big nothing will happen, if little no matter int or bytes they will be converted
        :return: TlSchema or None
        """
        if isinstance(tl_id, bytes):
            if byteorder == 'little':
                tl_id = tl_id[::-1]
        if isinstance(tl_id, int):
            tl_id = tl_id.to_bytes(4, byteorder)
        return self.id_map.get(tl_id, None)  # or TlSchema.empty()?

    def get_by_name(self, name: str) -> TlSchema:
        """
        :param name: name of TL Schema
        :return: TlSchema or None
        """
        return self.name_map.get(name, None)  # or TlSchema.empty()?

    def get_by_class_name(self, class_name: str) -> typing.List[TlSchema]:
        """
        :param class_name: boxed class_name of TL Schema
        :return: TlSchema or None
        """
        return self.class_name_map.get(class_name, None)  # or TlSchema.empty()?

    def generate_map(self):
        for schema in self.list:
            schema: TlSchema
            if schema.is_empty():
                continue
            self.id_map[schema.id] = schema
            self.name_map[schema.name] = schema
            self.class_name_map[schema.class_name] = self.class_name_map.get(schema.class_name, []) + [schema]

    def set_default_untouchables(self):
        self.untouchables['adnl.message.part'] = {'data'}

    def serialize_field(self, type_: str, value):
        logger.log(level=5, msg=f'serializing {type_} with value {value}')
        result = b''
        if type_ in self.base_types:
            byte_len = self.base_types.get(type_)
            if byte_len:
                if isinstance(value, bool):
                    if value:
                        result += b'\xb5ur\x99'  # booltrue
                    else:
                        result += b'7\x97y\xbc'  # boolfalse
                elif isinstance(value, bytes):
                    result += value[:byte_len][::-1] + b'\x00' * max(0, byte_len - len(value))
                elif isinstance(value, int):
                    result += value.to_bytes(length=byte_len, byteorder='little', signed=True)
                elif isinstance(value, str):
                    result += bytes.fromhex(value)
            else:
                if type_ == 'bytes':
                    if isinstance(value, dict) and '@type' in value:
                        value = self.serialize(schema=self.get_by_name(value['@type']), data=value, boxed=True)
                    if isinstance(value, bytes):
                        temp = b''
                        bytes_len = len(value)
                        if bytes_len <= 253:
                            temp += bytes_len.to_bytes(length=1, byteorder='little')
                        else:
                            temp += b'\xFE' + bytes_len.to_bytes(length=3, byteorder='little')
                        temp += value
                        if len(temp) % 4:
                            temp += (4 - len(temp) % 4) * b'\x00'
                        result += temp
        else:
            schemas = self.get_by_class_name(type_)
            if schemas:  # implicit
                if len(schemas) == 1:
                    result += self.serialize(schemas[0], value, boxed=True)
                else:
                    if isinstance(value, bytes):
                        result += value  # can specify implicit scheme value already serialized
                    elif isinstance(value, dict) and '@type' in value:
                        result += self.serialize(schema=self.get_by_name(value['@type']), data=value, boxed=True)
                    else:
                        raise TlError(f'Unknown value provided for implicit schemes {schemas}: {value}')
            else:  # explicit
                if type_.startswith('('):
                    subtype = type_.split()[1][:-1]
                    if 'vector' in type_:
                        temp = len(value).to_bytes(4, 'little', signed=False)
                        for v in value:
                            temp += self.serialize_field(subtype, v)
                        result += temp
                else:
                    result += self.serialize(self.get_by_name(type_), value, boxed=False)
        return result

    def serialize(self, schema: TlSchema, data: dict, boxed: bool = True) -> bytes:
        logger.log(level=5, msg=f'serializing schema {schema}')
        # https://core.telegram.org/mtproto/serialize
        """
        :param schema: TlSchema object
        :param data: {'key': value} - data to serialize
        :param boxed: need TL id prefix?
        :return: TL-serialized bytes
        """
        if boxed:
            result = schema.little_id()
        else:
            result = b''
        for field, type_ in schema.args.items():
            if 'mode' in type_ or 'flags' in type_:
                type_ = type_.split('?')[1]
                if data.get(field) is None:
                    continue
            value = data[field]
            result += self.serialize_field(type_, value)
        logger.log(level=5, msg=f'serialization result for schema {schema} is {result.hex()}')
        return result

    def deserialize(self, data: bytes, boxed: bool = True, args=None) -> typing.Tuple[typing.Union[dict, bytes], int]:
        i = 0
        result = {}
        schema = None
        if boxed:
            schema = self.get_by_id(data[i:i + 4], 'little')
            if not schema:  # is None
                return data, len(data)
            result['@type'] = schema.name
                # return {'bytes': data}, len(data)
            i += 4
            args = schema.args
        logger.log(level=5, msg=f'deserializing schema with args {args}')
        for field, type_ in args.items():
            if '?' in type_:
                index = int(type_[type_.find('.') + 1: type_.find('?')])
                mask = bin(result.get('mode', result.get('flags'))).replace('0b', '')[::-1]
                if index >= len(mask):
                    continue
                if mask[index] == '0':
                    continue
                type_ = type_.split('?')[-1]

            if type_ in self.base_types:
                byte_len = self.base_types.get(type_)
                if byte_len:  # is not None
                    if type_ == 'Bool':
                        if data[i:i + byte_len] == b'\xb5ur\x99':
                            result[field] = True
                        elif data[i:i + byte_len] == b'7\x97y\xbc':
                            result[field] = False
                    elif type_ in ('int128', 'int256'):
                        result[field] = data[i:i + byte_len].hex()
                    else:
                        result[field] = int.from_bytes(data[i:i + byte_len], 'little', signed=True)
                    i += byte_len
                else:
                    if type_ in ('bytes', 'string'):
                        if data[i:i+1] == b'\xFE':
                            # b'\xFE' means data len took more than one byte
                            byte_len = int.from_bytes(data[i+1:i+4], 'little')
                            attach_len = 4
                            i += 4
                        else:
                            attach_len = 1
                            byte_len = int.from_bytes(data[i:i+1], 'little')
                            i += 1
                        if ((not self._auto_deserialize) or
                                (schema is not None and
                                 schema.name in self.untouchables
                                 and field in self.untouchables[schema.name])):
                            result[field] = data[i:i + byte_len]
                        else:
                            temp, j = self.deserialize(data[i:i+byte_len])
                            if j < byte_len:
                                while j < byte_len:
                                    if field not in result:
                                        result[field] = [temp]
                                    temp, jj = self.deserialize(data[i + j:i + byte_len])
                                    j += jj
                                    if jj == 0:
                                        result[field] = data[i:i+byte_len]
                                        break
                                    result[field].append(temp)
                            else:
                                result[field] = temp
                        i += byte_len
                        if (byte_len + attach_len) % 4:
                            i += 4 - (byte_len + attach_len) % 4
                        if type_ == 'string':
                            result[field] = result[field].decode()
            else:
                # stucks when errors # TODO
                if type_.startswith('('):
                    subtype = type_.split()[1][:-1]
                    sch = self.get_by_name(subtype)
                    if 'vector' in type_:
                        length = int.from_bytes(data[i:i + 4], 'little', signed=False)
                        i += 4
                        result[field] = []
                        for _ in range(length):
                            if sch:
                                deser, j = self.deserialize(data[i:], False, sch.args)
                            else:
                                deser, j = self.deserialize(data[i:], True)

                            result[field].append(deser)
                            i += j
                else:
                    sch = self.get_by_name(type_)
                    if sch:
                        result[field], j = self.deserialize(data[i:], False, sch.args)
                        result[field]['@type'] = sch.name
                    else:
                        result[field], j = self.deserialize(data[i:], True)
                    # result[field], j = self.deserialize(data[i:], False, sch.args)
                    i += j
        return result, i

    def __repr__(self):
        return '[' + '\n'.join([i.__repr__() for i in self.list]) + ']'


def split(fields: str):
    result = {}
    temp = ''
    temp_key = ''
    br = 0
    for l in fields:
        if l == '(':
            temp += l
            br += 1
            continue
        if l == ')':
            temp += l
            br -= 1
            continue
        if l == ':':
            temp_key = temp
            temp = ''
            continue
        if l == ' ':
            if br == 0:
                if temp_key:
                    result[temp_key] = temp
                temp_key = ''
                temp = ''
            else:
                temp += l
            continue
        temp += l
    return result


class TlRegistrator:

    def __init__(self):
        self._re = re.compile(r"\s([^:]+):(\(.+\)|\S+)")
        self._base_types = TlSchemas.base_types
        # https://www.debuggex.com/r/kT4s0-gThkHLZCGO ; to avoid recursion in reg ex and use built-in "re" lib.
        # self._re = re.compile(r'[\w]*[^\w\s]*\[[^\]]*\]|'
        #                             r'[\w]*[^\w\s]*\([^)(]*(?:\([^)(]*(?:\([^)(]*(?:\([^)(]*\)[^)(]*)*\)[^)(]*)*\)[^)(]*)*\)|'
        #                             r'[\w]*[^\w\s]*\{[^\}]*\}|'
        #                             r'\S+')

    def _is_boxed(self, args: dict) -> bool:
        for type_ in args.values():
            if type_ not in self._base_types:
                return True
        return False

    def register(self, schema: str) -> TlSchema:
        schema = schema.split('//')[0]
        name = schema.split(' ')[0]
        if '#' in name:
            split_name = name.split('#')
            tl_id = bytes.fromhex(split_name[1])
            name = split_name[0]
        else:
            tl_id = self.get_id(schema.strip())
        # args = {i: j for i, j in self._re.findall(schema)}
        args = split(' '.join(schema.split()[1:-1]))
        class_name = schema.split(' ')[-1].replace(';', '')
        return TlSchema(tl_id, name, class_name, args)

    @staticmethod
    def clear(schema: str) -> str:
        return schema.replace(';', '').replace('(', '').replace(')', '')

    @staticmethod
    def crc32(schema: str) -> int:
        return zlib.crc32(schema.encode())

    def get_id(self, schema: str) -> bytes:
        return self.crc32(self.clear(schema)).to_bytes(4, 'big')

    def get_params(self, schema: str) -> dict:
        pass


class TlGenerator:
    def __init__(self, path: str, registrator: typing.Optional[TlRegistrator] = None) -> None:
        self._path = os.path.normpath(path)
        if registrator is None:
            registrator = TlRegistrator()
        self._registrator = registrator

    @classmethod
    def with_default_schemas(cls) -> "TlGenerator":
        path = os.path.join(os.path.dirname(__file__), './schemas')
        return cls(path)

    def generate(self):
        result = []
        if os.path.isdir(self._path):
            for f in os.listdir(self._path):
                result += self.from_file(os.path.join(self._path, f))
        else:
            result = self.from_file(self._path)
        return TlSchemas(result)

    def from_file(self, file_path: str):
        result = []
        with open(file_path, 'r') as f:
            temp = ''
            for line in f:
                stripped = line.strip()

                if not stripped or stripped.startswith('//') or stripped.startswith('---'):
                    continue
                if ';' not in stripped:
                    temp += stripped + ' '
                    continue
                else:
                    stripped = temp + stripped
                    temp = ''
                result.append(self._registrator.register(stripped))
        return result

